from .api import *  # noqa
from .api import __all__ as api_exports

# Delegate to API export
__all__ = api_exports

# Package metadata
__author__ = 'Tomas Aparicio'
__license__ = 'MIT'

# Current version
__version__ = '0.1.10'



# -*- coding: utf-8 -*-

import unittest
import requests
import pook


class TestUnitTestEngine(unittest.TestCase):

    @pook.activate
    def test_simple_pook_request(self):
        pook.get('server.com/foo').reply(204)
        res = requests.get('http://server.com/foo')
        self.assertEqual(res.status_code, 204)

    def test_enable_engine(self):
        pook.get('server.com/foo').reply(204)
        res = requests.get('http://server.com/foo')
        self.assertEqual(res.status_code, 204)

    @pook.get('server.com/foo', reply=204)
    def test_decorator(self):
        res = requests.get('http://server.com/foo')
        self.assertEqual(res.status_code, 204)

    def test_context_manager(self):
        with pook.use():
            pook.get('server.com/bar', reply=204)
            res = requests.get('http://server.com/bar')
            self.assertEqual(res.status_code, 204)

    def test_no_match_exception(self):
        pook.get('server.com/bar', reply=204)
        try:
            requests.get('http://server.com/baz')
        except:
            pass
        else:
            raise RuntimeError('expected to fail')



import functools
from copy import deepcopy
from abc import abstractmethod, ABCMeta
from ..compare import compare


class BaseMatcher(object):
    """
    BaseMatcher implements the basic HTTP request matching interface.
    """

    __metaclass__ = ABCMeta

    # Negate matching if necessary
    negate = False

    def __init__(self, expectation, negate=False):
        if not expectation:
            raise ValueError('expectation argument cannot be empty')

        self.negate = negate
        self._expectation = expectation

    @property
    def name(self):
        return type(self).__name__

    @property
    def expectation(self):
        return self._expectation

    @expectation.setter
    def expectation(self, value):
        self._expectation = value

    @abstractmethod
    def match(self, request):
        """
        Match performs the value matching.
        This is an abstract method that must be implemented by child classes.

        Arguments:
            request (pook.Request): request object to match.
        """
        pass

    def compare(self, value, expectation, regex_expr=False):
        """
        Compares two values with regular expression matching support.

        Arguments:
            value (mixed): value to compare.
            expectation (mixed): value to match.
            regex_expr (bool, optional): enables string based regex matching.

        Returns:
            bool
        """
        return compare(value, expectation, regex_expr=regex_expr)

    def to_dict(self):
        """
        Returns the current matcher representation as dictionary.

        Returns:
            dict
        """
        return {self.name: deepcopy(self.expectation)}

    def __repr__(self):
        return '{}({})'.format(self.name, self.expectation)

    def __str__(self):
        return self.expectation

    @staticmethod
    def matcher(fn):
        @functools.wraps(fn)
        def wrapper(self, *args):
            result = fn(self, *args)
            return not result if self.negate else result
        return wrapper



# MIME type aliases for semantic convenience
TYPES = {
  'text': 'text/plain',
  'html': 'text/html',
  'json': 'application/json',
  'xml': 'application/xml',
  'urlencoded': 'application/x-www-form-urlencoded',
  'form': 'application/x-www-form-urlencoded',
  'form-data': 'application/x-www-form-urlencoded'
}



import sys
from .base import BaseMatcher

if sys.version_info < (3,):     # Python 2
    from urlparse import parse_qs
else:                           # Python 3
    from urllib.parse import parse_qs


class QueryMatcher(BaseMatcher):
    """
    QueryMatcher implements an URL query params matcher.
    """

    def match_query(self, query, req_query):
        def test(key, param):
            match = req_query.get(key)
            if match is None:
                return False

            # Compare query params
            for index, value in enumerate(param):
                if index >= len(match):
                    return False
                # Perform query param comparison
                self.compare(value, match[index])

            return True

        return all([test(key, query[key]) for key in query])

    @BaseMatcher.matcher
    def match(self, req):
        query = self.expectation

        # Parse and assert type
        if isinstance(query, str):
            query = parse_qs(self.expectation)

        # Validate query params
        if not isinstance(query, dict):
            raise ValueError('query params must be a str or dict')

        # Parse request URL query
        req_query = parse_qs(req.url.query)

        # Match query params
        return self.match_query(query, req_query)



# -*- coding: utf-8 -*-

import re
from pook.regex import isregex_expr, isregex


def test_isregex_expr():
    cases = (
        ('re/[a-z]/', True),
        ('re/[0-9]/', True),
        ('re/[(.*)]/', True),
        ('re//', True),
        ('RE/[0-9]/', False),
        ('/[0-9]/', False),
        ('[0-9]', False),
        ('//', False),
        ('re/', False),
        ('re/[0-1]/-', False),
        ('e/[0-1]/-', False),
        ('e/[0-1]/', False),
        ('', False),
        ([], False),
        (1, False),
        (None, False),
    )

    for case in cases:
        assert isregex_expr(case[0]) is case[1]


def test_isregex():
    cases = (
        (re.compile('[a-z]'), True),
        ('[a-z]', False),
        ('', False),
        ([], False),
        (1, False),
        (None, False),
    )

    for case in cases:
        assert isregex(case[0]) is case[1]



import re
import sys
from unittest import TestCase
from .regex import isregex, strip_regex, isregex_expr

# If running Python 3
PY_3 = sys.version_info >= (3,)


def equal(x, y):
    """
    Shortcut function for ``unittest.TestCase.assertEqual()``.

    Arguments:
        x (mixed)
        y (mixed)

    Raises:
        AssertionError: in case of assertion error.

    Returns:
        bool
    """
    if PY_3:
        return TestCase().assertEqual(x, y) or True

    assert x == y


def matches(x, y, regex_expr=False):
    """
    Tries to match a regular expression value ``x`` against ``y``.
    Aliast``unittest.TestCase.assertEqual()``

    Arguments:
        x (regex|str): regular expression to test.
        y (str): value to match.
        regex_expr (bool): enables regex string based expression matching.

    Raises:
        AssertionError: in case of mismatching.

    Returns:
        bool
    """
    # Parse regex expression, if needed
    x = strip_regex(x) if regex_expr and isregex_expr(x) else x

    # Run regex assertion
    if PY_3:
        # Retrieve original regex pattern
        x = x.pattern if isregex(x) else x
        # Assert regular expression via unittest matchers
        return TestCase().assertRegexpMatches(y, x) or True

    # Primitive regex matching for Python 2.7
    if isinstance(x, str):
        x = re.compile(x, re.IGNORECASE)

    assert x.match(y) is not None


def test(x, y, regex_expr=False):
    """
    Compares to values based on regular expression matching or
    strict equality comparison.

    Arguments:
        x (regex|str): string or regular expression to test.
        y (str): value to match.
        regex_expr (bool): enables regex string based expression matching.

    Raises:
        AssertionError: in case of matching error.

    Returns:
        bool
    """
    return matches(x, y, regex_expr=regex_expr) if isregex(x) else equal(x, y)



# -*- coding: utf-8 -*-

import pytest
from pook import MockEngine, Engine
from pook.interceptors import BaseInterceptor


class Interceptor(BaseInterceptor):
    def activate(self):
        self.active = True

    def disable(self):
        self.active = False


@pytest.fixture
def engine():
    return MockEngine(Engine())


def test_mock_engine_instance(engine):
    assert isinstance(engine.engine, Engine)
    assert isinstance(engine.interceptors, list)
    assert len(engine.interceptors) >= 2


def test_mock_engine_flush(engine):
    assert len(engine.interceptors) >= 2
    engine.flush_interceptors()
    assert len(engine.interceptors) == 0


def test_mock_engine_interceptors(engine):
    engine.flush_interceptors()
    engine.add_interceptor(Interceptor)
    assert len(engine.interceptors) == 1
    assert isinstance(engine.interceptors[0], Interceptor)

    engine.remove_interceptor('Interceptor')
    assert len(engine.interceptors) == 0


def test_mock_engine_status(engine):
    engine.flush_interceptors()
    engine.add_interceptor(Interceptor)
    assert len(engine.interceptors) == 1

    interceptor = engine.interceptors[0]
    engine.activate()
    assert interceptor.active

    engine.disable()
    assert not interceptor.active



import json
from .decorators import fluent
from .headers import HTTPHeaderDict
from .helpers import trigger_methods
from .constants import TYPES


class Response(object):
    """
    Response is used to declare and compose an HTTP mock responses fields.

    It provides a chainable DSL interface for easier and declarative usage.

    Arguments:
        status (int): HTTP response status code. Defaults to ``200``.
        headers (dict): HTTP response headers.
        body (str|bytes): HTTP response body.
        json (str|dict|list): HTTP response JSON body.
        xml (str): HTTP response XML body.
        type (str): HTTP response content MIME type.
        file (str): file path to HTTP body response.

    Attributes:
        mock (pook.Mock): reference to mock instance.
    """

    def __init__(self, **kw):
        self._status = 200
        self._mock = None
        self._body = None
        self._headers = HTTPHeaderDict()

        # Trigger response method based on input arguments
        trigger_methods(self, kw)

    @fluent
    def status(self, code=200):
        """
        Defines the response status code.

        Arguments:
            code (int): response status code.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self._status = int(code)

    @fluent
    def header(self, key, value):
        """
        Defines a new response header.
        Alias to ``Response.header()``.

        Arguments:
            header (str): header name.
            value (str): header value.

        Returns:
            self: ``pook.Response`` current instance.
        """
        if type(key) is tuple:
            key, value = str(key[0]), key[1]

        headers = {key: value}
        self._headers.extend(headers)

    @fluent
    def headers(self, headers):
        """
        Defines a new response header.
        Alias to ``Response.header()``.

        Arguments:
            header (str): header name.
            value (str): header value.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self._headers.extend(headers)

    @fluent
    def set(self, header, value):
        """
        Defines a new response header.
        Alias to ``Response.header()``.

        Arguments:
            header (str): header name.
            value (str): header value.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self._headers[header] = value

    @fluent
    def type(self, name):
        """
        Defines the response ``Content-Type`` header.
        Alias to ``Response.content(mime)``.

        You can pass one of the following type aliases instead of the full
        MIME type representation:

        - ``json`` = ``application/json``
        - ``xml`` = ``application/xml``
        - ``html`` = ``text/html``
        - ``text`` = ``text/plain``
        - ``urlencoded`` = ``application/x-www-form-urlencoded``
        - ``form`` = ``application/x-www-form-urlencoded``
        - ``form-data`` = ``application/x-www-form-urlencoded``

        Arguments:
            value (str): type alias or header value to match.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self.content(name)

    @fluent
    def content(self, name):
        """
        Defines the response ``Content-Type`` header.

        You can pass one of the following type aliases instead of the full
        MIME type representation:

        - ``json`` = ``application/json``
        - ``xml`` = ``application/xml``
        - ``html`` = ``text/html``
        - ``text`` = ``text/plain``
        - ``urlencoded`` = ``application/x-www-form-urlencoded``
        - ``form`` = ``application/x-www-form-urlencoded``
        - ``form-data`` = ``application/x-www-form-urlencoded``

        Arguments:
            value (str): type alias or header value to match.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self._headers['Content-Type'] = TYPES.get(name, name)

    @fluent
    def body(self, body):
        """
        Defines response body data.

        Arguments:
            body (str|bytes): response body to use.

        Returns:
            self: ``pook.Response`` current instance.
        """
        if isinstance(body, bytes):
            body = body.decode('utf-8')

        self._body = body

    @fluent
    def json(self, data):
        """
        Defines the mock response JSON body.

        Arguments:
            data (dict|list|str): JSON body data.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self._headers['Content-Type'] = 'application/json'
        if not isinstance(data, str):
            data = json.dumps(data, indent=4)
        self._body = data

    @fluent
    def xml(self, xml):
        """
        Defines the mock response XML body.

        For not it only supports ``str`` as input type.

        Arguments:
            xml (str): XML body data to use.

        Returns:
            self: ``pook.Response`` current instance.
        """
        self.body(xml)

    @fluent
    def file(self, path):
        """
        Defines the response body from file contents.

        Arguments:
            path (str): disk file path to load.

        Returns:
            self: ``pook.Response`` current instance.
        """
        with open(path, 'r') as f:
            self.body = str(f.read())

    @property
    def mock(self):
        """
        Getter accessor for `mock` attribute.
        """
        return self._mock

    @mock.setter
    def mock(self, mock):
        """
        Setter for ``mock`` attribute.
        """
        self._mock = mock

    def __repr__(self):
        """
        Returns an human friendly readable instance data representation.

        Returns:
            str
        """
        args = []
        for key in ('headers', 'status', 'body'):
            value = getattr(self, '_{}'.format(key))
            args.append('{}={}'.format(key, value))
        return 'Response(\n    {}\n)'.format(',\n    '.join(args))



# -*- coding: utf-8 -*-

import pytest
from pook.mock import Mock


@pytest.fixture
def mock():
    return Mock()


def matcher(mock):
    return mock.matchers[0]


def test_mock_url(mock):
    mock.url('http://google.es')
    assert str(matcher(mock)) == 'http://google.es'



# -*- coding: utf-8 -*-

import pook
import pytest
import requests


@pook.activate
def test_simple_pook_request():
    pook.get('server.com/foo').reply(204)
    res = requests.get('http://server.com/foo')
    assert res.status_code == 204


def test_enable_engine():
    pook.get('server.com/foo').reply(204)
    res = requests.get('http://server.com/foo')
    assert res.status_code == 204
    pook.disable()


@pook.get('server.com/bar', reply=204)
def test_decorator():
    res = requests.get('http://server.com/bar')
    assert res.status_code == 204


def test_context_manager():
    with pook.use():
        pook.get('server.com/baz', reply=204)
        res = requests.get('http://server.com/baz')
        assert res.status_code == 204


def test_no_match_exception():
    pook.get('server.com/bar', reply=204)
    with pytest.raises(Exception):
        requests.get('http://server.com/baz')



# -*- coding: utf-8 -*-

import subprocess

# List of engine specific test commands to run
engine_tests = (
    'py.test tests/integration/engines/pytest_suite.py',
    'nosetests tests/integration/engines/nose_suite.py',
    'python -m unittest tests.integration.engines.unittest_suite',
)


def test_engines():
    for cmd in engine_tests:
        args = cmd.split(' ')
        code = subprocess.call(args)

        if code != 0:
            raise AssertionError('invalid exit code for: {}'.format(cmd))



# -*- coding: utf-8 -*-

import re
import pytest
from pook.request import Request
from pook.matchers.url import URLMatcher


def run_tests(cases, regex=False):
    for test in cases:
        match_url, url, matches = test
        req = Request(url=url)

        if regex:
            match_url = re.compile(match_url, re.IGNORECASE)

        if matches:
            assert URLMatcher(match_url).match(req)
        else:
            with pytest.raises(Exception):
                URLMatcher(match_url).match(req)


def test_url_matcher_urlparse():
    run_tests((
        # Valid cases
        ('http://foo.com', 'http://foo.com', True),
        ('http://foo.com:80', 'http://foo.com:80', True),
        ('http://foo.com', 'http://foo.com/foo/bar', True),
        ('http://foo.com/foo', 'http://foo.com/foo', True),
        ('http://foo.com/foo/bar', 'http://foo.com/foo/bar', True),
        ('http://foo.com/foo/bar/baz', 'http://foo.com/foo/bar/baz', True),
        ('http://foo.com/foo?x=y&z=w', 'http://foo.com/foo?x=y&z=w', True),

        # Invalid cases
        ('http://foo.com', 'http://bar.com', False),
        ('http://foo.com:80', 'http://foo.com:443', False),
        ('http://foo.com/foo', 'http://foo.com', False),
        ('http://foo.com/foo', 'http://foo.com/bar', False),
        ('http://foo.com/foo/bar', 'http://foo.com/bar/foo', False),
        ('http://foo.com/foo/bar/baz', 'http://foo.com/baz/bar/foo', False),
        ('http://foo.com/foo?x=y&z=w', 'http://foo.com/foo?x=x&y=y', False),
    ))


def test_url_matcher_regex():
    run_tests((
        # Valid cases
        ('http://foo.com', 'http://foo.com', True),
        ('http://foo.com:80', 'http://foo.com:80', True),
        ('^http://foo.com', 'http://foo.com/foo/bar', True),
        ('http://foo.com/foo', 'http://foo.com/foo', True),
        ('http://foo.com/foo/bar', 'http://foo.com/foo/bar', True),
        ('http://foo.com/foo/bar/baz', 'http://foo.com/foo/bar/baz', True),
        ('http://foo.com/foo\?x=[0-9]', 'http://foo.com/foo?x=5', True),

        # Invalid cases
        ('http://foo.com', 'http://bar.com', False),
        ('http://foo.com:80', 'http://foo.com:443', False),
        ('^http://foo.com$', 'http://foo.com/bar', False),
        ('http://foo.com/foo', 'http://foo.com/bar', False),
        ('http://foo.com/foo/bar', 'http://foo.com/bar/foo', False),
        ('http://foo.com/foo/bar/baz', 'http://foo.com/baz/bar/foo', False),
        ('http://foo.com/foo\?x=[1-3]', 'http://foo.com/foo?x=5', False),
    ), regex=True)



# -*- coding: utf-8 -*-

import pytest
from pook.matchers.base import BaseMatcher


class _BaseMatcher(BaseMatcher):

    def match(self, x):
        pass


def test_base_matcher_instance():
    matcher = _BaseMatcher('foo')
    assert matcher.name == '_BaseMatcher'
    assert matcher.negate is False
    assert matcher.expectation == 'foo'
    assert matcher.to_dict() == {'_BaseMatcher': 'foo'}
    assert matcher.__repr__() == '_BaseMatcher(foo)'
    assert matcher.__str__() == 'foo'


def test_base_matcher_compare():
    assert _BaseMatcher('foo').compare('foo', 'foo')

    assert _BaseMatcher('foo').compare('foo', 'foo')

    with pytest.raises(AssertionError):
        assert _BaseMatcher('foo').compare('foo', 'bar')


def test_base_matcher_exceptions():
    assert _BaseMatcher('foo').match(None) is None

    with pytest.raises(ValueError,
                       message='expectation argument cannot be empty'):
        _BaseMatcher(None)


def test_base_matcher_matcher():
    assert BaseMatcher.matcher(lambda x: True)(BaseMatcher)

    matcher = _BaseMatcher('foo', negate=True)
    assert BaseMatcher.matcher(lambda x: False)(matcher)



from inspect import ismethod, isfunction
from .exceptions import PookInvalidArgument


def trigger_methods(instance, args):
    """"
    Triggers specific class methods using a simple reflection
    mechanism based on the given input dictionary params.

    Arguments:
        instance (object): target instance to dynamically trigger methods.
        args (iterable): input arguments to trigger objects to
    """
    # Start the magic
    for name in sorted(args):
        value = args[name]
        target = instance

        # If response attibutes
        if name.startswith('response_') or name.startswith('reply_'):
            name = name.replace('response_', '').replace('reply_', '')
            # If instance has response attribute, use it
            if hasattr(instance, '_response'):
                target = instance._response

        # Retrieve class member for inspection and future use
        member = getattr(target, name, None)

        # Is attribute
        isattr = name in dir(target)
        iscallable = ismethod(member) and not isfunction(member)

        if not iscallable and not isattr:
            raise PookInvalidArgument('Unsupported argument: {}'.format(name))

        # Set attribute or trigger method
        if iscallable:
            member(value)
        else:
            setattr(target, name, value)



import sys
from .urllib3 import Urllib3Interceptor
from .http import HTTPClientInterceptor
from .base import BaseInterceptor

# Explicit symbols to export
__all__ = (
    'interceptors', 'add', 'get',
    'BaseInterceptor',
    'Urllib3Interceptor',
    'HTTPClientInterceptor',
    'AIOHTTPInterceptor',
)

# Store built-in interceptors in pook.
interceptors = [
    Urllib3Interceptor,
    HTTPClientInterceptor
]

# Import aiohttp in modern Python runtimes
if sys.version_info >= (3, 4, 2):
    from .aiohttp import AIOHTTPInterceptor
    interceptors.append(AIOHTTPInterceptor)


def add(*interceptors):
    """
    Registers a new HTTP client interceptor.

    Arguments:
        *interceptors (interceptor): interceptor(s) to be added.
    """
    interceptors.append(*interceptors)


def get(name):
    """
    Returns an interceptor by class name.

    Arguments:
        name (str): interceptor class name or alias.

    Returns:
        interceptor: found interceptor instance, otherwise ``None``.
    """
    for interceptor in interceptors:
        if interceptor.__name__ == name:
            return interceptor



import sys
from ..request import Request
from .base import BaseInterceptor

# Support Python 2/3
try:
    import mock
except:
    from unittest import mock

if sys.version_info < (3,):     # Python 2
    from urlparse import urlunparse
    from httplib import responses as http_reasons
else:                           # Python 3
    from urllib.parse import urlunparse
    from http.client import responses as http_reasons

if sys.version_info >= (3, 4, 2):  # Python 3.4.2+
    import asyncio
else:
    asyncio = None

# Try to load yarl URL parser package used by aiohttp
try:
    import yarl
    import multidict
except:
    yarl, multidict = None, None

PATCHES = (
    'aiohttp.client.ClientSession._request',
)

RESPONSE_CLASS = 'ClientResponse'
RESPONSE_PATH = 'aiohttp.client_reqrep'


def HTTPResponse(*args, **kw):
    # Dynamically load package
    module = __import__(RESPONSE_PATH, fromlist=(RESPONSE_CLASS,))
    ClientResponse = getattr(module, RESPONSE_CLASS)

    # Return response instance
    return ClientResponse(*args, **kw)


class AIOHTTPInterceptor(BaseInterceptor):
    """
    aiohttp HTTP client traffic interceptor.
    """

    def _url(self, url):
        return yarl.URL(url) if yarl else None

    @asyncio.coroutine
    def _on_request(self, _request, session, method, url,
                    data=None, headers=None, **kw):
        # Create request contract based on incoming params
        req = Request(method)
        req.headers = headers or {}
        req.body = data

        # Expose extra variadic arguments
        req.extra = kw

        # Compose URL
        req.url = str(url)

        # Match the request against the registered mocks in pook
        mock = self.engine.match(req)

        # If cannot match any mock, run real HTTP request if networking
        # or silent model are enabled, otherwise this statement won't
        # be reached (an exception will be raised before).
        if not mock:
            return _request(session, method, url,
                            data=data, headers=headers, **kw)

        # Simulate network delay
        if mock._delay:
            yield from asyncio.sleep(mock._delay / 1000)  # noqa

        # Shortcut to mock response
        res = mock._response

        # Aggregate headers as list of tuples for interface compatibility
        headers = []
        for key in res._headers:
            headers.append((key, res._headers[key]))

        # Create mock equivalent HTTP response
        _res = HTTPResponse(req.method, self._url(urlunparse(req.url)))

        # response status
        _res.version = (1, 1)
        _res.status = res._status
        _res.reason = http_reasons.get(res._status)
        _res._should_close = False

        # Add response headers
        _res.raw_headers = tuple(headers)
        _res.headers = multidict.CIMultiDictProxy(
            multidict.CIMultiDict(headers)
        )

        # Define `_content` attribute with an empty string to
        # force do not read from stream (which won't exists)
        _res._content = ''
        if res._body:
            _res._content = res._body.encode('utf-8', errors='replace')

        # Return response based on mock definition
        return _res

    def _patch(self, path):
        # If not modern Python, just ignore patch
        if not asyncio:
            return None

        @asyncio.coroutine
        def handler(session, method, url, data=None, headers=None, **kw):
            return (yield from self._on_request(
                _request, session, method, url,
                data=data, headers=headers, **kw)
            )

        try:
            # Create a new patcher for Urllib3 urlopen function
            # used as entry point for all the HTTP communications
            patcher = mock.patch(path, handler)
            # Retrieve original patched function that we might need for real
            # networking
            _request = patcher.get_original()[0]
            # Start patching function calls
            patcher.start()
        except:
            # Exceptions may accur due to missing package
            # Ignore all the exceptions for now
            pass
        else:
            self.patchers.append(patcher)

    def activate(self):
        """
        Activates the traffic interceptor.
        This method must be implemented by any interceptor.
        """
        [self._patch(path) for path in PATCHES]

    def disable(self):
        """
        Disables the traffic interceptor.
        This method must be implemented by any interceptor.
        """
        [patch.stop() for patch in self.patchers]



import sys
import json as _json
from .headers import HTTPHeaderDict
from .helpers import trigger_methods

if sys.version_info < (3,):     # Python 2
    from urlparse import urlparse, parse_qs, urlunparse
else:                           # Python 3
    from urllib.parse import urlparse, parse_qs, urlunparse


class Request(object):
    """
    Request object representing the request mock expectation DSL.

    Arguments:
        method (str): HTTP method to match. Defaults to ``GET``.
        url (str): URL request to intercept and match.
        headers (dict): HTTP headers to match.
        query (dict): URL query params to match. Complementely to URL
            defined query params.
        body (str|regex): request body payload to match.
        json (str|dict|list): JSON payload body structure to match.
        xml (str): XML payload data structure to match.

    Attributes:
        method (str): HTTP method to match. Defaults to ``GET``.
        url (str): URL request to intercept and match.
        headers (dict): HTTP headers to match.
        query (dict): URL query params to match. Complementely to URL
            defined query params.
        body (str|regex): request body payload to match.
        json (str|dict|list): JSON payload body structure to match.
        xml (str): XML payload data structure to match.
    """

    # Store keys
    keys = ('method', 'headers', 'body', 'url', 'query')

    def __init__(self, method='GET', **kw):
        self._url = None
        self._body = None
        self._query = None
        self._method = method
        self._extra = kw.get('extra')
        self._headers = HTTPHeaderDict()

        trigger_methods(self, kw)

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        self._method = method

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, headers):
        if not hasattr(headers, '__setitem__'):
            raise TypeError('headers must be a dictionary')
        self._headers.extend(headers)

    @property
    def extra(self):
        return self._extra

    @extra.setter
    def extra(self, extra):
        if not isinstance(extra, dict):
            raise TypeError('extra must be a dictionary')
        self._extra = extra

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = urlparse(url)
        self._query = parse_qs(self._url.query)

    @property
    def query(self, url):
        return self._query

    @query.setter
    def query(self, params):
        self._query = parse_qs(params)

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, body):
        if hasattr(body, 'decode'):
            try:
                body = body.decode('utf-8', 'strict')
            except:
                pass

        self._body = body

    @property
    def json(self):
        return _json.loads(self._body)

    @json.setter
    def json(self, data):
        if isinstance(data, str):
            self._body = data
        else:
            self._body = _json.dumps(data)

    @property
    def xml(self):
        return self._body

    @xml.setter
    def xml(self, data):
        self._body = data

    def copy(self):
        """
        Copies the current Request object instance for side-effects purposes.

        Returns:
            pook.Request: copy of the current Request instance.
        """
        req = type(self)()
        req.__dict__ = self.__dict__.copy()
        req._headers = self.headers.copy()
        return req

    def __repr__(self):
        """
        Returns an human friendly readable instance data representation.

        Returns:
            str
        """
        args = []
        for key in self.keys:
            value = getattr(self, '_{}'.format(key))
            if key == 'url':
                value = urlunparse(value)
            args.append('{}={},\n'.format(key, value))
        return 'Request(\n  {})'.format('  '.join(args))



from .base import BaseMatcher


class HeadersMatcher(BaseMatcher):
    """
    Headers HTTP request matcher.
    """

    def __init__(self, headers):
        if not isinstance(headers, dict):
            raise TypeError('headers must be a dictionary')
        BaseMatcher.__init__(self, headers)

    @BaseMatcher.matcher
    def match(self, req):
        for key in self.expectation:
            # Retrieve value to match
            value = self.expectation[key]

            # Retrieve header value by key
            header = req.headers.get(key)

            # Compare header value
            print('Match:', self.compare(value, header, regex_expr=True))
            if not self.compare(value, header, regex_expr=True):
                return False

        return True



import json
from .base import BaseMatcher


class JSONMatcher(BaseMatcher):
    """
    JSONMatcher implements a JSON body matcher supporting strict structure
    and regular expression based comparisons.
    """

    def __init__(self, data):
        BaseMatcher.__init__(self, data)

        if isinstance(data, str):
            self.expectation = json.loads(data)

    @BaseMatcher.matcher
    def match(self, req):
        body = req.body

        if isinstance(body, str):
            try:
                body = json.loads(body)
            except:
                return False

        x = json.dumps(self.expectation, sort_keys=True, indent=4)
        y = json.dumps(body, sort_keys=True, indent=4)

        return self.compare(x, y)



import functools


def fluent(fn):
    """
    Simple function decorator allowing easy method chaining.

    Arguments:
        fn (function): target function to decorate.
    """
    @functools.wraps(fn)
    def wrapper(self, *args, **kw):
        # Trigger method proxy
        result = fn(self, *args, **kw)
        # Return self instance or method result
        return self if result is None else result
    return wrapper



# -*- coding: utf-8 -*-

import re
import os
import subprocess

# Regular expression used to match Python files
pyfile = re.compile('.py$')

# List of allowed example to fail
allowed_errors = (
    'simulated_error.py'
)


def test_examples():
    # List of file examples
    examples = [f for f in os.listdir('examples') if pyfile.match(f)]

    # Test file example
    for example in examples:
        code = subprocess.call(['python', 'examples/{}'.format(example)])

        expected = 1 if example in allowed_errors else 0
        if code != expected:
            raise AssertionError('invalid exit code: {}'.format(example))
